""" Objects pertaining to the various ingestion manifests """

from __future__ import annotations

from enum import Enum


# pylint: disable=E0401, R0903, R1721, W0622


class Telescope(Enum):
    """Codifying names of telescopes that appear in ingestion manifests"""

    VLA = "VLA"
    EVLA = "EVLA"
    ALMA = "ALMA"
    VLBA = "VLBA"
    GBT = "GBT"
    NONE = "NONE"
    UNKNOWN = "UNKNOWN"


class ScienceProductType(Enum):
    """Canonical collection of ingestible types of science products"""

    EXEC_BLOCK = "execution_block"
    EVLA_CAL = "calibration"
    CATALOG = "catalog"
    IMAGE = "image"


class AncillaryProductType(Enum):
    """The various types of ancillary products we'll encounter"""

    INGESTION_ARTIFACTS = "ingestion_artifacts"
    PIPELINE_ARTIFACTS = "pipeline_artifacts"
    PIPELINE_WEBLOG = "pipeline_weblog"
    LOG = "log_file"

    ### Images ###

    # our default FITS type
    FITS = "fits_image"

    VLASS_QUICKLOOK = "quicklook_rms_image"
    AUDI_FITS_MASK = "clean_mask"
    AUDI_PB_FITS = "primary_beam"
    ALPHA_FITS = "spectral_index"
    CANDIDATE_IMG = "candidate_image"
    THUMBNAIL_IMG = "thumbnail_image"


class InvalidLocatorException(Exception):
    """Throw this if we're fed a bad science product locator"""


class IngestionManifestException(Exception):
    """Throw this if we're unable to construct an ingestion manifest using supplied inputs"""


def find_output_science_products(files_found, staging_source_dir):
    """
    Round up the output science products associated with this SP type.

    :return:
    """
    sp_files = [file for file in files_found if file.name.endswith(".tar")]
    if len(sp_files) == 0:
        raise IngestionManifestException(
            f"No output science products found at {staging_source_dir}"
        )

    return sp_files
